/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "rollball.hpp"
#include "fggl/data/model.hpp"
#include "fggl/data/procedural.hpp"
#include "fggl/gfx/camera.hpp"
#include "fggl/input/camera_input.h"

struct Prefabs {
	fggl::ecs3::entity_t wallX;
	fggl::ecs3::entity_t wallZ;
	fggl::ecs3::entity_t floor;
	fggl::ecs3::entity_t collectable;
	fggl::ecs3::entity_t player;
};

static void setupPrefabs(fggl::ecs3::World& world, Prefabs& prefabs) {

	{
		// walls
		prefabs.wallX = world.create(true);
		world.add<fggl::math::Transform>(prefabs.wallX);

		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";

		// create a procedural cube, but scale the model to be a long wall
		auto transform = glm::scale(fggl::data::OFFSET_NONE, {1.0f, 5.f, 41.0f});
		fggl::data::make_cube(meshComponent.mesh, transform );
		world.set<fggl::data::StaticMesh>(prefabs.wallX, &meshComponent);

		auto* rb = world.add<fggl::phys::RigidBody>(prefabs.wallX);
		rb->mass = fggl::phys::MASS_STATIC; // flag the object as static
		rb->shape = new fggl::phys::Box({0.5F, 2.5F, 20.5F});
	}

	{
		// walls
		prefabs.wallZ = world.create(true);
		world.add<fggl::math::Transform>(prefabs.wallZ);

		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";

		// create a procedural cube, but scale the model to be a long wall
		auto transform = glm::scale(fggl::data::OFFSET_NONE, {39.f, 5.f, 1.0f});
		fggl::data::make_cube(meshComponent.mesh, transform );
		world.set<fggl::data::StaticMesh>(prefabs.wallZ, &meshComponent);

		auto* rb = world.add<fggl::phys::RigidBody>(prefabs.wallZ);
		rb->mass = fggl::phys::MASS_STATIC; // flag the object as static
		rb->shape = new fggl::phys::Box({39.0F / 2, 2.5F, 0.5F});
	}

	{
		// floor
		prefabs.floor = world.create(true);
		world.add<fggl::math::Transform>(prefabs.floor);

		// create a procedural cube, but scale it to be plane-like
		// TODO proper planes
		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";
		auto transform = glm::scale(fggl::data::OFFSET_NONE, {40.f, 0.5f, 40.0f});
		fggl::data::make_cube(meshComponent.mesh, transform );
		world.set<fggl::data::StaticMesh>(prefabs.floor, &meshComponent);

		// physics: a static cube
		auto* rb = world.add<fggl::phys::RigidBody>(prefabs.floor);
		rb->mass = fggl::phys::MASS_STATIC; // flag the object as static
		rb->shape = new fggl::phys::Box({20.0F, 0.5F, 20.0F});
	}

	{
		// player (cube because my sphere function doesn't exist yet
		prefabs.player = world.create(true);
		world.add<fggl::math::Transform>(prefabs.player);
		world.add<fggl::phys::Dynamics>(prefabs.player);
		world.add<fggl::phys::RigidBody>(prefabs.player);

		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";
		fggl::data::make_sphere(meshComponent.mesh);
		world.set<fggl::data::StaticMesh>(prefabs.player, &meshComponent);
	}

	{
		// collectable
		prefabs.collectable = world.create(true);
		world.add<fggl::math::Transform>(prefabs.collectable);

		fggl::data::StaticMesh meshComponent;
		meshComponent.pipeline = "phong";
		fggl::data::make_cube(meshComponent.mesh);
		world.set<fggl::data::StaticMesh>(prefabs.collectable, &meshComponent);

		auto* callbacks = world.add<fggl::phys::CollisionCallbacks>(prefabs.collectable);
	}

}

static void setupCamera(fggl::ecs3::World& world) {
	auto prototype = world.create(false);

	// setup camera position/transform
	auto* transform = world.add<fggl::math::Transform>(prototype);
	if ( transform != nullptr) {
		transform->origin(glm::vec3(10.0f, 3.0f, 10.0f));
	}

	// setup camera components
	world.add<fggl::gfx::Camera>(prototype);
}

static fggl::ecs3::entity_t setupEnvironment(fggl::ecs3::World& world, const Prefabs prefabs, fggl::math::vec2& size) {
	{
		auto northWall = world.copy(prefabs.wallX);
		auto* transform = world.get<fggl::math::Transform>(northWall);
		transform->origin({size.x/2, 0.0f, 0.0f});
	}

	{
		auto southWall = world.copy(prefabs.wallX);
		auto* transform = world.get<fggl::math::Transform>(southWall);
		transform->origin({-size.x/2, 0.0f, 0.0f});
	}

	{
		auto westWall = world.copy(prefabs.wallZ);
		auto* transform = world.get<fggl::math::Transform>(westWall);
		transform->origin({0.0f, 0.0f, -size.y/2});
	}

	{
		auto eastWall = world.copy(prefabs.wallZ);
		auto* transform = world.get<fggl::math::Transform>(eastWall);
		transform->origin({0.0f, 0.0f, size.y/2});
	}

	{
		auto floor = world.copy(prefabs.floor);
		auto *transform = world.get<fggl::math::Transform>(floor);
		transform->origin({0.0f, -2.5f, 0.0f});
	}

	auto player = fggl::ecs3::NULL_ENTITY;
	{
		// player just starts off as the prefab dictates
		player = world.copy(prefabs.player);
	}

	{
		// collectables
		std::array<fggl::math::vec3, 3> collectPos {{
			{-5.0f, -0.5f, 12.0f},
			{15.0f, -0.5f, 0.5f},
			{6.0f, -0.5f, -15.0f}
		}};

		// build the collectables
		for (auto& pos : collectPos) {
			auto collectable = world.copy(prefabs.collectable);
			auto* transform = world.get<fggl::math::Transform>(collectable);
			transform->origin(pos);
		}
	}

	return player;
}


namespace demo {

	RollBall::RollBall(fggl::App &app) : Game(app) {
	}

	void RollBall::activate() {
		Game::activate();

		Prefabs prefabs{};
		setupPrefabs(world(), prefabs);

		// actual scene objects
		setupCamera(world());

		// create a 20x20 grid
		fggl::math::vec2 size{40.0f, 40.0f};
		player = setupEnvironment(world(), prefabs, size);
	}

	void RollBall::update() {
		Game::update();
		auto& input = this->input();

			if ( player != fggl::ecs3::NULL_ENTITY ) {
				auto &world = this->world();

				const fggl::math::vec3 forward = fggl::math::FORWARD;
				const fggl::math::vec3 right = fggl::math::RIGHT;
				fggl::math::vec3 force = fggl::math::VEC3_ZERO;
				bool moved = false;

				if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_W))) {
					force += forward;
					moved = true;
				}

				if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_S))) {
					force -= forward;
					moved = true;
				}

				if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_A))) {
					force -= right;
					moved = true;
				}

				if (input.keyboard.down(glfwGetKeyScancode(GLFW_KEY_D))) {
					force += right;
					moved = true;
				}

				// setup dynamic force
				if ( moved ) {
					float forceFactor = 3.0F;
					force = glm::normalize(force) * forceFactor;
					auto *dynamics = world.get<fggl::phys::Dynamics>(player);
					dynamics->force += force;
				}

				{
					auto cameras = world.findMatching<fggl::gfx::Camera>();
					fggl::ecs3::entity_t cam = cameras[0];
					auto *camComp = world.get<fggl::gfx::Camera>(cam);
					camComp->target = world.get<fggl::math::Transform>(player)->origin();

					auto *camTransform = world.get<fggl::math::Transform>(cam);
					camTransform->origin( camComp->target - (fggl::math::FORWARD * 15.0F) + (fggl::math::UP * 15.0F) );
				}

		}
	}

	void RollBall::render(fggl::gfx::Graphics &gfx) {
		Game::render(gfx);
	}

} // namespace demo

